part of worksdata;

/// Works action type, this is to identify different type of action log.
enum WorksActionType {
  BUTTON,
  START_PAGE,
  END_PAGE,
  EVENT,
  DYNAMIC_LINK,
  PUSH_NOTIFICATION,
  TAB,
  PAUSED_PAGE,
  RESUMED_PAGE,
  CAMPAIGN,
}

/// Environment ID (pointing to different database)
enum WorksdataEnv {
  DEV,
  QA,
  PROD,
}

/// A configuration class. Pass in this class as configuration.
///
/// token: A secret token used to communicate with API.
/// env: 'DEV', 'QA', 'PROD', determine where to log.
class WorksdataConfig {
  final String? token;
  final WorksdataEnv? env;

  WorksdataConfig({this.token, this.env});

  Map toJson() => {
        'token': token,
        'env': env,
      };
}

/// Main class for the Worksdata package.
class Worksdata {
  static final Worksdata _worksdata = Worksdata._internal();

  factory Worksdata() {
    return _worksdata;
  }

  Worksdata._internal();

  final Uuid _uuid = Uuid();
  final String _platform = Platform.operatingSystem.toUpperCase();
  ApiService _api = ApiService(api: API.worksdataDev());
  final deviceInfoPlugin = DeviceInfoPlugin();
  late SharedPreferences prefs;
  String? _tempId;
  String? _token;
  String? _campaign;
  Map? _campaignData;
  String? _countryCode;
  Map? _configuration;

  /// To initiate Works plugin. Ensure this method is called in the main.dart or
  /// the place initiating the application.
  ///
  Future<void> init({required WorksdataConfig worksdataConfig}) async {
    prefs = await SharedPreferences.getInstance();
    _token = worksdataConfig.token ?? '';
    _tempId = prefs.getString('worksdata_temp_id');
    _setupConfiguration(worksdataConfig);
    if (_tempId == null || _tempId!.isEmpty) {
      await _setTempId();
    }
    await _setCountryCode();
    log('SYSTEM_INIT', worksActionType: WorksActionType.EVENT);
  }

  /// Send log information to Works API
  ///
  /// --actionId: log identifier to determine what log is this. (String with UPPER_SNAKE_CASE)
  /// --data: extra data to this log.
  /// --type: log type. eg: EVENT, BUTTON, DYNAMIC_LINK (Prefer using upper snake case)
  /// --worksActionType: replace 'type' if defined. Use predefined 'WorksActionType'.
  ///
  void log(String actionId,
      {Map? data, String? type, WorksActionType? worksActionType}) async {
    if (_countryCode == null) {
      await _setCountryCode();
    }
    if (worksActionType != null &&
        worksActionType == WorksActionType.CAMPAIGN) {
      await prefs.setString('worksdata_campaign', actionId);
      await prefs.setString('worksdata_campaign_data', jsonEncode(data));
    }
    _campaign = prefs.getString('worksdata_campaign');
    _campaignData =
        jsonDecode(prefs.getString('worksdata_campaign_data') ?? 'null');
    var actionLog = {
      'id': actionId.toUpperCase(),
      'type': worksActionType != null
          ? worksActionType.toString().split('.').last
          : (type ?? '').toUpperCase(),
      'created_time': DateTime.now().millisecondsSinceEpoch,
      'temp_id': _tempId,
      'platform': _platform,
      'data': data,
      'campaign': _campaign,
      'campaign_data': _campaignData,
      'additional_information': {
        'country_code': _countryCode,
      },
    };
    unawaited(_api.post(
      APIPath.log(),
      body: {'data': actionLog, 'configuration': _configuration},
      header: {'Authorization': 'Bearer $_token'},
    ));
  }

  /// Get uid & sending mapping between uid and tempId to Works API
  ///
  /// --uid: user's ID. Usually can be retrieved when logging in the user.
  /// --toLog: to log or not this event(login)
  ///
  void login(String uid, {bool toLog = true, Map? additionalInfo}) {
    if (toLog) log('SYSTEM_LOGIN', worksActionType: WorksActionType.EVENT);
    _api.get(
      APIPath.login(),
      queryParameters: {
        'env': _configuration!['env'],
        'uid': uid,
        'temp_id': _tempId,
        'additional_info': jsonEncode(additionalInfo),
      },
      header: {'Authorization': 'Bearer $_token'},
    );
  }

  /// Flush uid & reset the tempId.
  ///
  /// --toLog: to log or not this event(logout)
  ///
  Future<void> logout({bool toLog = true}) async {
    await _setTempId();
    if (toLog) log('SYSTEM_LOGOUT', worksActionType: WorksActionType.EVENT);
  }

  /// Create a new temporary id and store in Shared Preference.
  Future<void> _setTempId() async {
    _tempId = await _getTempId();
    await prefs.setString('worksdata_temp_id', _tempId!);
  }

  /// Get the temporary id from user device id and uuid v4.
  Future<String> _getTempId() async =>
      '${await _getUserDeviceId()}${_uuid.v4()}';

  /// Get user device id.
  Future<String?> _getUserDeviceId() async {
    if (Platform.isAndroid) {
      return (await deviceInfoPlugin.androidInfo).androidId;
    } else if (Platform.isIOS) {
      return (await deviceInfoPlugin.iosInfo).identifierForVendor;
    }
    return null;
  }

  /// Set _countryCode
  Future<void> _setCountryCode() async {
    _countryCode ??= await _getCountryFromGeolocation();
    _countryCode ??= _getCountryFromIsoCode();
  }

  /// Get country code from geolocation.
  Future<String?> _getCountryFromGeolocation() async {
    var geolocationResult = (await _getGeolocation());
    if (geolocationResult != null) {
      return geolocationResult['alpha2'];
    }
    return null;
  }

  /// Get country code from locale.
  String? _getCountryFromIsoCode() => window.locale.countryCode;

  /// API that getting geolocation.
  Future<Map?> _getGeolocation() async =>
      await (ApiService(api: API.ipGeolocation())
          .get(await _getUserIpAddress())
          .catchError((e) {
        return null;
      }) as FutureOr<Map<dynamic, dynamic>?>);

  /// Getting user's network ip address.
  Future<String?> _getUserIpAddress() async {
    final _api = ApiService(api: API.ipify());
    var ipAddress = await _api.get(APIPath.ipifyIpAddress()).catchError((e) {});
    return ipAddress != null ? ipAddress['ip'] : null;
  }

  /// Re-format configuration which used to store or communicate with API.
  void _setupConfiguration(WorksdataConfig _config) {
    var formattedConfiguration = _config.toJson();
    if (_config.env is WorksdataEnv) {
      formattedConfiguration['env'] = _config.env.toString().split('.').last;
      switch (_config.env) {
        case WorksdataEnv.DEV:
          _api = ApiService(api: API.worksdataDev());
          break;
        case WorksdataEnv.QA:
          _api = ApiService(api: API.worksdataQA());
          break;
        case WorksdataEnv.PROD:
          _api = ApiService(api: API.worksdataProd());
          break;
      }
    }
    _configuration = formattedConfiguration;
  }
}
